@@ -1,9 +1,10 @@
 #!/usr/bin/env python
 # Copyright (c) 2012 Eric S. Raymond <esr@thyrsus.com>
-# Distributed under BSD terms.
+# SPDX-License-Identifier: BSD-2-Clause
 #
 # This script contains git porcelain and porcelain byproducts.
-# Requires Python 2.6, or 2.5 with the simplejson library installed.
+# Requires either Python 2.6, or 2.5 with the simplejson library installed
+# or Python 3.x.
 #
 # usage: irkerhook.py [-V] [-n] [--variable=value...] [commit_id...]
 #
@@ -13,15 +14,20 @@
 #
 # See the irkerhook manual page in the distribution for a detailed
 # explanation of how to configure this hook.
+from __future__ import print_function
 
 # The default location of the irker proxy, if the project configuration
 # does not override it.
+#
+# SPDX-License-Identifier: BSD-2-Clause
+from __future__ import print_function, absolute_import
+
 default_server = "localhost"
 IRKER_PORT = 6659
 
 # The default service used to turn your web-view URL into a tinyurl so it
 # will take up less space on the IRC notification line.
-default_tinyifier = "http://tinyurl.com/api-create.php?url="
+default_tinyifier = u"http://tinyurl.com/api-create.php?url="
 
 # Map magic urlprefix values to actual URL prefixes.
 urlprefixmap = {
@@ -30,24 +36,45 @@ urlprefixmap = {
     "cgit": "http://%(host)s/cgi-bin/cgit.cgi/%(repo)s/commit/?id=",
     }
 
-# By default, ship to the freenode #commits list 
-default_channels = "irc://chat.freenode.net/#commits"
+# By default, ship to the freenode #commits list
+default_channels = u"irc://chat.freenode.net/#commits"
 
 #
 # No user-serviceable parts below this line:
 #
 
-version = "1.17"
+version = "2.19"
 
-import os, sys, commands, socket, urllib, subprocess, locale
+import os, sys, socket, subprocess, locale, datetime, re
 from pipes import quote as shellquote
+
+try:
+    from urllib2 import urlopen, HTTPError
+except ImportError:
+    from urllib.error import HTTPError
+    from urllib.request import urlopen
+
 try:
     import simplejson as json	# Faster, also makes us Python-2.5-compatible
 except ImportError:
     import json
 
+if sys.version_info.major == 2:
+    string_type = unicode
+else:
+    string_type = str
+
+try:
+    getstatusoutput = subprocess.getstatusoutput
+except AttributeError:
+    import commands
+    getstatusoutput = commands.getstatusoutput
+
 def do(command):
-    return unicode(commands.getstatusoutput(command)[1], locale.getlocale()[1] or 'UTF-8').encode(locale.getlocale()[1] or 'UTF-8')
+    if sys.version_info.major == 2:
+        return string_type(getstatusoutput(command)[1], locale.getlocale()[1] or 'UTF-8')
+    else:
+        return getstatusoutput(command)[1]
 
 class Commit:
     def __init__(self, extractor, commit):
@@ -60,22 +87,31 @@ class Commit:
         self.files = None
         self.logmsg = None
         self.url = None
+        self.author_date = None
+        self.commit_date = None
         self.__dict__.update(extractor.__dict__)
-    def __unicode__(self):
+
+        if sys.version_info.major == 2:
+            # Convert __str__ to __unicode__ for python 2
+            self.__unicode__ = self.__str__
+            # Not really needed, but maybe useful for debugging
+            self.__str__ = lambda x: x.__unicode__().encode('utf-8')
+
+    def __str__(self):
         "Produce a notification string from this commit."
         if self.urlprefix.lower() == "none":
             self.url = ""
         else:
-            urlprefix = urlprefixmap.get(self.urlprefix, self.urlprefix) 
+            urlprefix = urlprefixmap.get(self.urlprefix, self.urlprefix)
             webview = (urlprefix % self.__dict__) + self.commit
             try:
-                if urllib.urlopen(webview).getcode() == 404:
-                    raise IOError
+                # See it the url is accessible
+                res = urlopen(webview)
                 if self.tinyifier and self.tinyifier.lower() != "none":
                     try:
-                        # Didn't get a retrieval error or 404 on the web
+                        # Didn't get a retrieval error on the web
                         # view, so try to tinyify a reference to it.
-                        self.url = open(urllib.urlretrieve(self.tinyifier + webview)[0]).read()
+                        self.url = urlopen(self.tinyifier + webview).read()
                         try:
                             self.url = self.url.decode('UTF-8')
                         except UnicodeError:
@@ -84,19 +120,28 @@ class Commit:
                         self.url = webview
                 else:
                     self.url = webview
+            except HTTPError as e:
+                if e.code == 401:
+                    # Authentication error, so we assume the view is valid
+                    self.url = webview
+                else:
+                    self.url = ""
             except IOError:
                 self.url = ""
-        return self.template % self.__dict__
+        res = self.template % self.__dict__
+        return string_type(res, 'UTF-8') if not isinstance(res, string_type) else res
 
 class GenericExtractor:
     "Generic class for encapsulating data from a VCS."
     booleans = ["tcp"]
     numerics = ["maxchannels"]
+    strings = ["email"]
     def __init__(self, arguments):
         self.arguments = arguments
         self.project = None
         self.repo = None
         # These aren't really repo data but they belong here anyway...
+        self.email = None
         self.tcp = True
         self.tinyifier = default_tinyifier
         self.server = None
@@ -132,7 +177,7 @@ class GenericExtractor:
             self.bold = '\x1b[1m'
             self.green = '\x1b[1;32m'
             self.blue = '\x1b[1;34m'
-            self.red =  '\x1b[1;31m'
+            self.red = '\x1b[1;31m'
             self.yellow = '\x1b[1;33m'
             self.brown = '\x1b[33m'
             self.magenta = '\x1b[35m'
@@ -181,6 +226,8 @@ class GenericExtractor:
                     setattr(self, key, False)
             elif key in GenericExtractor.numerics:
                 setattr(self, key, int(val))
+            elif key in GenericExtractor.strings:
+                setattr(self, key, val)
         if not self.project:
             sys.stderr.write("irkerhook.py: no project name set!\n")
             raise SystemExit(1)
@@ -193,7 +240,11 @@ class GenericExtractor:
 
 def has(dirname, paths):
     "Test for existence of a list of paths."
-    return all([os.path.exists(os.path.join(dirname, x)) for x in paths])
+    # all() is a python2.5 construct
+    for exists in [os.path.exists(os.path.join(dirname, x)) for x in paths]:
+        if not exists:
+            return False
+    return True
 
 # VCS-dependent code begins here
 
@@ -211,11 +262,12 @@ class GitExtractor(GenericExtractor):
         self.repo = do("git config --get irker.repo")
         self.server = do("git config --get irker.server")
         self.channels = do("git config --get irker.channels")
+        self.email = do("git config --get irker.email")
         self.tcp = do("git config --bool --get irker.tcp")
-        self.template = '%(bold)s%(project)s:%(reset)s %(green)s%(author)s%(reset)s %(repo)s:%(yellow)s%(branch)s%(reset)s * %(bold)s%(rev)s%(reset)s / %(bold)s%(files)s%(reset)s: %(logmsg)s %(brown)s%(url)s%(reset)s'
+        self.template = do("git config --get irker.template") or u'%(bold)s%(project)s:%(reset)s %(green)s%(author)s%(reset)s %(repo)s:%(yellow)s%(branch)s%(reset)s * %(bold)s%(rev)s%(reset)s / %(bold)s%(files)s%(reset)s: %(logmsg)s %(brown)s%(url)s%(reset)s'
         self.tinyifier = do("git config --get irker.tinyifier") or default_tinyifier
         self.color = do("git config --get irker.color")
-        self.urlprefix = do("git config --get irker.urlprefix") or "gitweb"
+        self.urlprefix = do("git config --get irker.urlprefix") or u"gitweb"
         self.cialike = do("git config --get irker.cialike")
         self.filtercmd = do("git config --get irker.filtercmd")
         # These are git-specific
@@ -247,7 +299,7 @@ class GitExtractor(GenericExtractor):
     def commit_factory(self, commit_id):
         "Make a Commit object holding data for a specified commit ID."
         commit = Commit(self, commit_id)
-        commit.branch = os.path.basename(self.refname)
+        commit.branch = re.sub(r"^refs/[^/]*/", "", self.refname)
         # Compute a description for the revision
         if self.revformat == 'raw':
             commit.rev = commit.commit
@@ -256,21 +308,26 @@ class GitExtractor(GenericExtractor):
         else: # self.revformat == 'describe'
             commit.rev = do("git describe %s 2>/dev/null" % shellquote(commit.commit))
         if not commit.rev:
-            commit.rev = commit.commit[:12]
+            # Query git for the abbreviated hash
+            commit.rev = do("git log -1 '--pretty=format:%h' " + shellquote(commit.commit))
+            if self.urlprefix in ('gitweb', 'cgit'):
+                # Also truncate the commit used for the announced urls
+                commit.commit = commit.rev
         # Extract the meta-information for the commit
         commit.files = do("git diff-tree -r --name-only " + shellquote(commit.commit))
         commit.files = " ".join(commit.files.strip().split("\n")[1:])
-        # Design choice: for git we ship only the first line, which is
+        # Design choice: for git we ship only the first message line, which is
         # conventionally supposed to be a summary of the commit.  Under
         # other VCSes a different choice may be appropriate.
-        metainfo = do("git log -1 '--pretty=format:%an <%ae>|%s' " + shellquote(commit.commit))
-        (commit.author, _, commit.logmsg) = metainfo.partition("|")
-        commit.mail = commit.author.split()[-1].strip("<>")
+        commit.author_name, commit.mail, commit.logmsg = \
+            do("git log -1 '--pretty=format:%an%n%ae%n%s' " + shellquote(commit.commit)).split("\n")
         # This discards the part of the author's address after @.
         # Might be be nice to ship the full email address, if not
         # for spammers' address harvesters - getting this wrong
         # would make the freenode #commits channel into harvester heaven.
         commit.author = commit.mail.split("@")[0]
+        commit.author_date, commit.commit_date = \
+            do("git log -1 '--pretty=format:%ai|%ci' " + shellquote(commit.commit)).split("|")
         return commit
 
 class SvnExtractor(GenericExtractor):
@@ -299,6 +356,7 @@ class SvnExtractor(GenericExtractor):
         commit.branch = ""
         commit.rev = "r%s" % self.id
         commit.author = self.svnlook("author")
+        commit.commit_date = self.svnlook("date").partition('(')[0]
         commit.files = self.svnlook("dirs-changed").strip().replace("\n", " ")
         commit.logmsg = self.svnlook("log").strip()
         return commit
@@ -341,8 +399,9 @@ class HgExtractor(GenericExtractor):
         self.repo = ui.config('irker', 'repo')
         self.server = ui.config('irker', 'server')
         self.channels = ui.config('irker', 'channels')
+        self.email = ui.config('irker', 'email')
         self.tcp = str(ui.configbool('irker', 'tcp'))  # converted to bool again in do_overrides
-        self.template = '%(bold)s%(project)s:%(reset)s %(green)s%(author)s%(reset)s %(repo)s:%(yellow)s%(branch)s%(reset)s * %(bold)s%(rev)s%(reset)s / %(bold)s%(files)s%(reset)s: %(logmsg)s %(brown)s%(url)s%(reset)s'
+        self.template = ui.config('irker', 'template') or '%(bold)s%(project)s:%(reset)s %(green)s%(author)s%(reset)s %(repo)s:%(yellow)s%(branch)s%(reset)s * %(bold)s%(rev)s%(reset)s / %(bold)s%(files)s%(reset)s: %(logmsg)s %(brown)s%(url)s%(reset)s'
         self.tinyifier = ui.config('irker', 'tinyifier') or default_tinyifier
         self.color = ui.config('irker', 'color')
         self.urlprefix = (ui.config('irker', 'urlprefix') or
@@ -369,19 +428,29 @@ class HgExtractor(GenericExtractor):
         commit.rev = '%d:%s' % (ctx.rev(), commit.commit)
         commit.branch = ctx.branch()
         commit.author = person(ctx.user())
+        commit.author_date = \
+            datetime.datetime.fromtimestamp(ctx.date()[0]).strftime('%Y-%m-%d %H:%M:%S')
         commit.logmsg = ctx.description()
         # Extract changed files from status against first parent
         st = self.repository.status(ctx.p1().node(), ctx.node())
-        commit.files = ' '.join(st[0] + st[1] + st[2])
+        commit.files = ' '.join(st.modified + st.added + st.removed)
         return commit
 
 def hg_hook(ui, repo, **kwds):
-    # To be called from a Mercurial "commit" or "incoming" hook.  Example
-    # configuration:
+    # To be called from a Mercurial "commit", "incoming" or "changegroup" hook.
+    # Example configuration:
     # [hooks]
     # incoming.irker = python:/path/to/irkerhook.py:hg_hook
     extractor = HgExtractor([(ui, repo)])
-    ship(extractor, kwds['node'], False)
+    start = repo[kwds['node']].rev()
+    end = len(repo)
+    if start != end:
+        # changegroup with multiple commits, so we generate a notification
+        # for each one
+        for rev in range(start, end):
+            ship(extractor, rev, False)
+    else:
+        ship(extractor, kwds['node'], False)
 
 # The files we use to identify a Subversion repo might occur as content
 # in a git or hg repo, but the special subdirectories for those are more
@@ -390,6 +459,10 @@ extractors = [GitExtractor, HgExtractor, SvnExtractor]
 
 # VCS-dependent code ends here
 
+def convert_message(message):
+    """Convert the message to bytes to send to the socket"""
+    return message.encode(locale.getlocale()[1] or 'UTF-8') + b'\n'
+
 def ship(extractor, commit, debug):
     "Ship a notification for the specified commit."
     metadata = extractor.commit_factory(commit)
@@ -397,13 +470,13 @@ def ship(extractor, commit, debug):
     # This is where we apply filtering
     if extractor.filtercmd:
         cmd = '%s %s' % (shellquote(extractor.filtercmd),
-                         shellquote(json.dumps(metadata.__dict__)))
+                          shellquote(json.dumps(metadata.__dict__)))
         data = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE).stdout.read()
         try:
             metadata.__dict__.update(json.loads(data))
         except ValueError:
             sys.stderr.write("irkerhook.py: could not decode JSON: %s\n" % data)
-            raise SystemExit, 1
+            raise SystemExit(1)
 
     # Rewrite the file list if too long. The objective here is only
     # to be easier on the eyes.
@@ -422,10 +495,10 @@ def ship(extractor, commit, debug):
     # purposes the commit text is more important.  If it's still too long
     # there's nothing much can be done other than ship it expecting the IRC
     # server to truncate.
-    privmsg = unicode(metadata)
+    privmsg = string_type(metadata)
     if len(privmsg) > 510:
         metadata.files = ""
-        privmsg = unicode(metadata)
+        privmsg = string_type(metadata)
 
     # Anti-spamming guard.  It's deliberate that we get maxchannels not from
     # the user-filtered metadata but from the extractor data - means repo
@@ -437,23 +510,38 @@ def ship(extractor, commit, debug):
     # Ready to ship.
     message = json.dumps({"to": channels, "privmsg": privmsg})
     if debug:
-        print message
+        print(message)
     elif channels:
         try:
-            if extractor.tcp:
+            if extractor.email:
+                # We can't really figure out what our SF username is without
+                # exploring our environment. The mail pipeline doesn't care
+                # about who sent the mail, other than being from sourceforge.
+                # A better way might be to simply call mail(1)
+                sender = "irker@users.sourceforge.net"
+                msg = """From: %(sender)s
+Subject: irker json
+
+%(message)s""" % {"sender":sender, "message":message}
+                import smtplib
+                smtp = smtplib.SMTP()
+                smtp.connect()
+                smtp.sendmail(sender, extractor.email, msg)
+                smtp.quit()
+            elif extractor.tcp:
                 try:
                     sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                     sock.connect((extractor.server or default_server, IRKER_PORT))
-                    sock.sendall(message + "\n")
+                    sock.sendall(convert_message(message))
                 finally:
                     sock.close()
             else:
                 try:
                     sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
-                    sock.sendto(message + "\n", (extractor.server or default_server, IRKER_PORT))
+                    sock.sendto(convert_message(message), (extractor.server or default_server, IRKER_PORT))
                 finally:
                     sock.close()
-        except socket.error, e:
+        except socket.error as e:
             sys.stderr.write("%s\n" % e)
 
 if __name__ == "__main__":
@@ -464,7 +552,7 @@ if __name__ == "__main__":
         if arg == '-n':
             notify = False
         elif arg == '-V':
-            print "irkerhook.py: version", version
+            print("irkerhook.py: version", version)
             sys.exit(0)
         elif arg.startswith("--repository="):
             repository = arg[13:]
@@ -487,4 +575,7 @@ if __name__ == "__main__":
     for commit in commits:
         ship(extractor, commit, not notify)
 
-#End
+# The following sets edit modes for GNU EMACS
+# Local Variables:
+# mode:python
+# End:
